/* mlvalues.h */

/* structure, representation des valeurs */
/* En principe, ce fichier regroupe tout ce qui depend du modele memoire */

#ifndef _mlvalues_
#define _mlvalues_

#include "config.h"

typedef long value;
typedef value * address;

#define ADDRESS(obj) ((address) ((char *) (obj) - 4)) /* adresse du header */
#define OBJECT(adr)  ((value) ((char *) (adr) + 4))

/* Codage des entiers */

#define IS_INT(x)    (((x) & 1) == 1)
#define IS_BLOCK(x)  (((x) & 1) == 0)

#define MLINT(x)     (((long)(x) << 1) + 1)
#define CINT(x)      ((x) >> 1)
#define MAX_MLINT (0x3fffffff)
#define MIN_MLINT (-0x40000000)

/* Structure des blocs dans le tas */

/* Pour definir l'en-tete des blocs & les macros correspondantes */

typedef unsigned long header;
typedef unsigned long blocksize;

#define SIZE_HEADER 1

#define WHITE (0 << 8)
#define BLACK (3 << 8)
#define GRAY (1 << 8)
#define BLUE (2 << 8)

#define IS_WHITE(ptr) (COLOUR (ptr) == WHITE)
#define IS_GRAY(ptr) (COLOUR (ptr) == GRAY)
#define IS_BLACK(ptr) (COLOUR (ptr) == BLACK)
#define IS_BLUE(ptr) (COLOUR (ptr) == BLUE)


#define WHITE_TO_BLACK(ptr) { HEADER (ptr) |= BLACK; }
#define WHITE_TO_GRAY(ptr) { HEADER (ptr) |= GRAY; }
#define GRAY_TO_BLACK(ptr) { HEADER (ptr) |= BLUE; }
#define GRAY_TO_WHITE(ptr) { HEADER (ptr) &= ~GRAY; }
#define WHITE_TO_BLUE(ptr) { HEADER (ptr) |= BLUE; }
#define BLACK_TO_BLUE(ptr) { HEADER (ptr) &= ~GRAY; }
#define BLACK_TO_WHITE(ptr) { HEADER (ptr) &= ~BLACK; }
#define ANY_TO_WHITE(ptr) { HEADER (ptr) &= ~BLACK; }


/* Tags dans bits de poids faibles */
#define TAG_HDR(hd) ((hd) & 0xFF)
#define SIZE_HDR(hd) ((hd) >> 10)
#define COLOUR_HDR(hd) ((hd) & BLACK)

#define MAKE_HEADER(size, tag) \
       ((header) (((blocksize) (size) << 10) + (tag)))
  /* acces a l'en-tete */
#define HEADER(ptr) (*(header *) ((char*) (ptr) - 4))

#define NBR_TAG (1 << 8)	  /* le nombre de tags disponibles. */
#define MAX_TAG (NBR_TAG - 1)     /* la valeur maxi pour un tag */
#define MAX_SIZE ((1 << 22) - 1)  /* la taille maxi d'un bloc alloue */

  /* taille (en mots) d'un objet avec son header */
#define SIZE(ptr) (SIZE_HDR (HEADER (ptr)))
  /* taille (en mots) d'un objet sans header */
#define OBJECT_SIZE(ptr) ((SIZE_HDR (HEADER (ptr)) - 1)
#ifdef BIG_ENDIAN
#define TAG(ptr) (*((unsigned char *)(ptr) - 1))
#else
#define TAG(ptr) (*((unsigned char *)(ptr) - 4))
#endif
#define COLOUR(ptr) (COLOUR_HDR (HEADER (ptr)))

/* Le premier tag des valeurs qu'on ne suit pas au GC */
#define NO_SCAN_TAG (NBR_TAG - 3)

/* 1- Si tag < NO_SCAN_TAG : un n-uplet de valeurs.  */

  /* pointeur vers le premier champ */
#define POINTER(x) ((value *) (x))
  /* ieme champ (i = 0 => premier champ) */
#define FIELD(x, i) (((value *)(x)) [i])

#define CAR(x) (FIELD ((x), 0))
#define CDR(x) (FIELD ((x), 1))

/* Cas particulier: les fermetures */

typedef unsigned char instruction;
typedef instruction * code;

#define SIZE_CLOSURE 2
#define HEADER_CLOSURE MAKE_HEADER(SIZE_CLOSURE, 0)

#define CODE(x) (((code *)(x))[0])
#define ENV(x)  (FIELD(x, 1))


/* 2- Si tag >= NO_SCAN_TAG : une suite d'octets */

#define TAG_STRING NO_SCAN_TAG

  /* Retourne le ieme octet d'un objet */
#define BYTE(x,i) (((unsigned char *) (x)) [i])
#define CSTRING(x) (&(BYTE ((x), 0)))

  /* Trouve un bit dans un tableau de bits (retourne un entier 1 ou 0) */
#define BIT(x, i) ((BYTE (x, i / 8) >> (i % 8)) & 1)

/* Cas particulier: les flottants */

#define TAG_FLOAT (NO_SCAN_TAG + 1)
#define SIZE_FLOAT ((sizeof(double) / sizeof(value)))
#define HEADER_FLOAT MAKE_HEADER(SIZE_FLOAT, TAG_FLOAT)

/* Retourne la valeur flottante d'un objet */
#define CFLOAT(x) (* (double *) (x))

/* fonctions d'acces a partir d'un pointeur decode sur le header */

#define H_HEADER(p) (*(header *) p)
#define H_VALUE(p) (&(((header *) p) [1]))
#define H_FIELD(p, i) (((value *) H_VALUE (p)) [i])
#define H_SIZE(p) (SIZE_HDR (H_HEADER (p)))
#define H_TAG(p) (TAG_HDR (H_HEADER (p)))

/* pour obtenir la longueur d'une chaine ML dans une fonction C */
#define GET_STRING_LENGTH(s, size)					      \
{ blocksize temp = SIZE (s) << 2;					      \
  size = temp - BYTE (s, temp - 1);					      \
}

/* Pour les blocs non structures de mots de 32 bits */

#define TAG_BITVECT (NO_SCAN_TAG + 2)

/* Pour les blocs libres --- doit etre >= NO_SCAN_TAG pour shift_heap */

#define TAG_JUNK TAG_BITVECT

/* Cas special pour les blocs de taille nulle (atomes) */

extern header first_atoms[];

#define atom(tag) ((value) (first_atoms + 1 + (tag)))

#endif /* _mlvalues_ */
