#open "misc";;

let lex_half_buffer_size = 1024;;
let lex_buffer_size = 2 * lex_half_buffer_size;;
let lex_dummy_action () =
  (fatal_error "lexer: empty token" : int);;
let lex_buffer = create_string lex_buffer_size
and lex_aux_buffer = create_string lex_half_buffer_size
and lex_inchan = ref std_in
and lex_abs_pos = ref 0
and lex_curr_pos = ref lex_buffer_size
and lex_start_pos = ref lex_buffer_size
and lex_last_pos = ref lex_buffer_size
and lex_last_action = ref lex_dummy_action
;;

let init_lexer ic =
  lex_inchan := ic;
  lex_abs_pos := - lex_buffer_size;
  lex_curr_pos := lex_buffer_size;
  lex_start_pos := lex_buffer_size;
  lex_last_pos := lex_buffer_size;
  lex_last_action := lex_dummy_action
;;

let lex_refill () =
  let read = input (!lex_inchan) lex_aux_buffer 0 lex_half_buffer_size in
  let n =
    if read > 0 then read else (set_nth_char lex_aux_buffer 0 `\000`; 1)
  in
    blit_string lex_buffer n lex_buffer 0 (lex_buffer_size - n);
    blit_string lex_aux_buffer 0 lex_buffer (lex_buffer_size - n) n;
    lex_abs_pos := !lex_abs_pos + n;
    lex_curr_pos := !lex_curr_pos - n;
    lex_start_pos := !lex_start_pos - n;
    lex_last_pos := !lex_last_pos - n;
    if !lex_start_pos < 0 then fatal_error "lexer: token too long"
;;  

let get_next_char () =
  (if !lex_curr_pos >= lex_buffer_size then lex_refill ());
  nth_char lex_buffer (pred (incr lex_curr_pos))
;;

let get_lexeme_char i =
  nth_char lex_buffer (!lex_start_pos + i)
;;

let get_lexeme () =
  let len = !lex_curr_pos - !lex_start_pos in
  let s = create_string len in
    blit_string lex_buffer (!lex_start_pos) s 0 len;
    s
;;

let start_lexing () =
  lex_start_pos := !lex_curr_pos;
  lex_last_action := lex_dummy_action
;;

let backtrack () =
  lex_curr_pos := !lex_last_pos;
  (!lex_last_action) ()
;;

let start_token () =
  !lex_abs_pos + !lex_start_pos
and end_token () =
  !lex_abs_pos + !lex_curr_pos
;;

let reset_lexer_abs_pos () =
  lex_abs_pos := - !lex_curr_pos
;;
